/*	File gen.c: 2.1 (83/03/20,16:02:06) */
/*% cc -O -c %
 *
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "code.h"
#include "primary.h"
#include "sym.h"

static char *needargs[] = {
	"vreg",
	"vsync",
	"spr_hide", "spr_show",
	"satb_update",
	"map_load_tile",
	NULL
};

/*
 *	gen arg count
 *
 */
void gnargs (char *name, int nb)
{
	char *ptr;
	int   i;

	if (name == NULL)
		return;
	for (i = 0; ; i++) {
		ptr = needargs[i];

		if (ptr == NULL)
			break;
		if (strcmp(name, ptr) == 0) {
			out_ins(I_NARGS, T_VALUE, (DATA){.integer=nb});
			break;
		}
	}
}

/*
 *	return next available internal label number
 *
 */
int getlabel (void )
{
	return (nxtlab++);
}

/*
 *	fetch a static memory cell into the primary register
 */
void getmem (char *sym)
{
	if ((sym[IDENT] != POINTER) & (sym[TYPE] == CCHAR)) {
		if ((sym[STORAGE] & ~WRITTEN) == LSTATIC)
			out_ins(I_LDB, T_LABEL, (DATA){.integer=glint(sym)});
		else
			out_ins(I_LDB, T_SYMBOL, (DATA){.ptr=(sym + NAME)});
	} else {
		if ((sym[STORAGE] & ~WRITTEN) == LSTATIC)
			out_ins(I_LDW, T_LABEL, (DATA){.integer=glint(sym)});
		else
			out_ins(I_LDW, T_SYMBOL, (DATA){.ptr=(sym + NAME)});
	}
}

/*
 *	fetch a hardware register into the primary register
 */
void getio (char *sym)
{
	out_ins(I_CALL, T_LIB, (DATA){.ptr="getvdc"});
}
void getvram (char *sym)
{
	out_ins(I_CALL, T_LIB, (DATA){.ptr="readvram"});
}

/*
 *	fetch the address of the specified symbol into the primary register
 *
 */
void getloc (char *sym)
{
	int value;

	if ((sym[STORAGE] & ~WRITTEN) == LSTATIC)
		out_ins(I_LDWI, T_LABEL, (DATA){.integer=glint(sym)});
	else {
		value = glint(sym) - stkp;
		out_ins_sym(X_LEA_S, T_STACK, (DATA){.integer=value}, sym);

//		out_ins(I_LDW, T_STACK, NULL);
//
//		if (value || optimize) {
//			out_ins(I_ADDWI, T_VALUE, value);
//		}
	}
}

/*
 *	store the primary register into the specified static memory cell
 *
 */
void putmem (char *sym)
{
	int code;

	if ((sym[IDENT] != POINTER) & (sym[TYPE] == CCHAR))
		code = I_STB;
	else
		code = I_STW;

	if ((sym[STORAGE] & ~WRITTEN) == LSTATIC)
		out_ins(code, T_LABEL, (DATA){.integer=glint(sym)});
	else
		out_ins(code, T_SYMBOL, (DATA){.ptr=(sym + NAME)});
}

/*
 *	store the specified object type in the primary register
 *	at the address on the top of the stack
 *
 */
void putstk (char typeobj)
{
	if (typeobj == CCHAR)
		out_ins(I_STBPS, 0, (DATA){.ptr=NULL});
	else
		out_ins(I_STWPS, 0, (DATA){.ptr=NULL});
	stkp = stkp + INTSIZE;
}

/*
 *	store the primary register
 *	at the address on the top of the stack
 *
 */
void putio (char *sym)
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="setvdc"});
	stkp = stkp + INTSIZE;
}
void putvram (char *sym)
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="writevram"});
	stkp = stkp + INTSIZE;
}

/*
 *	fetch the specified object type indirect through the primary
 *	register into the primary register
 *
 */
void indirect (type_t typeobj)
{
	out_ins(I_STW, T_PTR, (DATA){.ptr=NULL});
	if (typeobj == CCHAR)
		out_ins(I_LDBP, T_PTR, (DATA){.ptr=NULL});
	else {
		out_ins(I_LDWP, T_PTR, (DATA){.ptr=NULL});
	}
}

void farpeek(char *ptr)
{
	if (ptr[TYPE] == CCHAR)
		out_ins(I_FGETB, T_SYMBOL, (DATA){.ptr=ptr});
	else
		out_ins(I_FGETW, T_SYMBOL, (DATA){.ptr=ptr});
}

/*
 *	print partial instruction to get an immediate value into
 *	the primary register
 *
 */
void immed (int type, DATA data)
{
	out_ins(I_LDWI, type, data);
}

/*
 *	push the primary register onto the stack
 *
 */
void gpush (void )
{
//	out_ins(I_PUSHWZ, T_VALUE, zpstkp);
//	zpstkp = zpstkp - INTSIZE;

	out_ins(I_PUSHW, T_VALUE, (DATA){.integer=INTSIZE});
	stkp = stkp - INTSIZE;
}

/*
 *	push the primary register onto the stack
 *
 */
void gpusharg (int size)
{
	out_ins(I_PUSHW, T_SIZE, (DATA){.integer=size});
	stkp = stkp - size;
}

/*
 *	pop the top of the stack into the secondary register
 *
 */
void gpop (void )
{
	out_ins(I_POPW, 0, (DATA){.ptr=NULL});
	stkp = stkp + INTSIZE;
}

/*
 *	swap the primary register and the top of the stack
 *
 */
void swapstk (void )
{
	out_ins(I_SWAPW, 0, (DATA){.ptr=NULL});
}

/*
 *	call the specified subroutine name
 *
 */
void gcall (char *sname, int nargs)
{
	out_ins_ex(I_CALL, T_SYMBOL, (DATA){.ptr=sname}, nargs);
}

/*
 *         generate a bank pseudo instruction
 *
 */
void gbank(unsigned char bank, unsigned short offset)
{
	out_ins(I_BANK, T_VALUE, (DATA){.integer=bank});
	out_ins(I_OFFSET, T_VALUE, (DATA){.integer=offset});
}

/*
 *	return from subroutine
 *
 */
void gret (void )
{
	out_ins(I_RTS, 0, (DATA){.ptr=NULL});
}

/*
 *	perform subroutine call to value on top of stack
 *
 */
void callstk (int nargs)
{
	if (nargs <= INTSIZE)
		out_ins(I_CALLS, T_STACK, (DATA){.integer=0});
	else
		out_ins(I_CALLS, T_STACK, (DATA){.integer=nargs - INTSIZE});

	stkp = stkp + INTSIZE;
}

/*
 *	jump to specified internal label number
 *
 */
void jump (int label)
{
	out_ins(I_LBRA, T_LABEL, (DATA){.integer=label});
}

/*
 *	test the primary register and jump if false to label
 *
 */
void testjump (int label, int ft)
{
	out_ins(I_TSTW, 0, (DATA){.ptr=NULL});
	if (ft)
		out_ins(I_LBNE, T_LABEL, (DATA){.integer=label});
	else
		out_ins(I_LBEQ, T_LABEL, (DATA){.integer=label});
}

/*
 *	modify the stack pointer to the new value indicated
 *      Is it there that we decrease the value of the stack to add local vars ?
 */
int modstk (int newstkp)
{
	int	k;

//	k = galign(newstkp - stkp);
	k = newstkp - stkp;
	if (k) {
		gtext();
		out_ins(I_ADDMI, T_STACK, (DATA){.integer=k});
	}
	return (newstkp);
}

/*
 *	multiply the primary register by INTSIZE
 */
void gaslint (void )
{
	out_ins(I_ASLW, 0, (DATA){.ptr=NULL});
}

/*
 *	divide the primary register by INTSIZE
 */
void gasrint(void )
{
	out_ins(I_ASRW, 0, (DATA){.ptr=NULL});
}

/*
 *	Case jump instruction
 */
void gjcase(void )
{
	out_ins(I_JMP, T_SYMBOL, (DATA){.ptr="__case"});
}

/*
 *	add the primary and secondary registers
 *	if lval2 is int pointer and lval is int, scale lval
 */
void gadd (expr_data_t *lval, expr_data_t *lval2)
{
	if (dbltest (lval2, lval)) {
		out_ins(I_ASLWS, 0, (DATA){.ptr=NULL});
	}
	out_ins(I_ADDWS, 0, (DATA){.ptr=NULL});
	stkp = stkp + INTSIZE;
}

/*
 *	subtract the primary register from the secondary
 *
 */
void gsub (void )
{
	out_ins(I_SUBWS, 0, (DATA){.ptr=NULL});
	stkp = stkp + INTSIZE;
}

/*
 *	multiply the primary and secondary registers
 *	(result in primary)
 *
 */
void gmult (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="smul"});
	stkp = stkp + INTSIZE;
}

/*
 *	divide the secondary register by the primary
 *	(quotient in primary, remainder in secondary)
 *
 */
void gdiv (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="sdiv"});
	stkp = stkp + INTSIZE;
}

/*
 *	compute the remainder (mod) of the secondary register
 *	divided by the primary register
 *	(remainder in primary, quotient in secondary)
 *
 */
void gmod (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="smod"});
	stkp = stkp + INTSIZE;
}

/*
 *	inclusive 'or' the primary and secondary registers
 *
 */
void gor (void )
{
	out_ins(I_ORWS, 0, (DATA){.ptr=NULL});
	stkp = stkp + INTSIZE;
}

/*
 *	exclusive 'or' the primary and secondary registers
 *
 */
void gxor (void )
{
	out_ins(I_EORWS, 0, (DATA){.ptr=NULL});
	stkp = stkp + INTSIZE;
}

/*
 *	'and' the primary and secondary registers
 *
 */
void gand (void )
{
	out_ins(I_ANDWS, 0, (DATA){.ptr=NULL});
	stkp = stkp + INTSIZE;
}

/*
 *	arithmetic shift right the secondary register the number of
 *	times in the primary register
 *	(results in primary register)
 *
 */
void gasr (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="asr"});
	stkp = stkp + INTSIZE;
}

/*
 *	arithmetic shift left the secondary register the number of
 *	times in the primary register
 *	(results in primary register)
 *
 */
void gasl (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="asl"});
	stkp = stkp + INTSIZE;
}

/*
 *	two's complement of primary register
 *
 */
void gneg (void )
{
	out_ins(I_NEGW, 0, (DATA){.ptr=NULL});
}

/*
 *	one's complement of primary register
 *
 */
void gcom (void )
{
	out_ins(I_COMW, 0, (DATA){.ptr=NULL});
}

/*
 *	convert primary register into logical value
 *
 */
void gbool (void )
{
	out_ins(I_BOOLW, 0, (DATA){.ptr=NULL});
}

/*
 *	logical complement of primary register
 *
 */
void glneg (void )
{
	out_ins(I_NOTW, 0, (DATA){.ptr=NULL});
}

/*
 *	increment the primary register by 1 if char, INTSIZE if
 *      int
 */
void ginc (expr_data_t * lval)
/* int lval[]; */
{
	if (lval->ptr_type == CINT)
		out_ins(I_ADDWI, T_VALUE, (DATA){.integer=2});
	else
		out_ins(I_ADDWI, T_VALUE, (DATA){.integer=1});
}

/*
 *	decrement the primary register by one if char, INTSIZE if
 *	int
 */
void gdec (expr_data_t *lval)
/* int lval[]; */
{
	if (lval->ptr_type == CINT)
		out_ins(I_SUBWI, T_VALUE, (DATA){.integer=2});
	else
		out_ins(I_SUBWI, T_VALUE, (DATA){.integer=1});
}

/*
 *	following are the conditional operators.
 *	they compare the secondary register against the primary register
 *	and put a literl 1 in the primary if the condition is true,
 *	otherwise they clear the primary register
 *
 */

/*
 *	equal
 *
 */
void geq (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="eq"});
	stkp = stkp + INTSIZE;
}

/*
 *	not equal
 *
 */
void gne (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="ne"});
	stkp = stkp + INTSIZE;
}

/*
 *	less than (signed)
 *
 */
/*void glt (int lvl) */
void glt (void)
{
	out_ins(I_JSR,  T_LIB, (DATA){.ptr="lt"});
	stkp = stkp + INTSIZE;
}

/*
 *	less than or equal (signed)
 *
 */
void gle (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="le"});
	stkp = stkp + INTSIZE;
}

/*
 *	greater than (signed)
 *
 */
void ggt (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="gt"});
	stkp = stkp + INTSIZE;
}

/*
 *	greater than or equal (signed)
 *
 */
void gge (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="ge"});
	stkp = stkp + INTSIZE;
}

/*
 *	less than (unsigned)
 *
 */
/*void gult (int lvl)*/
void gult (void)
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="ult"});
	stkp = stkp + INTSIZE;
}

/*
 *	less than or equal (unsigned)
 *
 */
void gule (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="ule"});
	stkp = stkp + INTSIZE;
}

/*
 *	greater than (unsigned)
 *
 */
void gugt (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="ugt"});
	stkp = stkp + INTSIZE;
}

/*
 *	greater than or equal (unsigned)
 *
 */
void guge (void )
{
	out_ins(I_JSR, T_LIB, (DATA){.ptr="uge"});
	stkp = stkp + INTSIZE;
}
