/*	File stmt.c: 2.1 (83/03/20,16:02:17) */
/*% cc -O -c %
 *
 */

#include <stdio.h>
#include "defs.h"
#include "data.h"
#include "code.h"
#include "error.h"
#include "expr.h"
#include "gen.h"
#include "io.h"
#include "lex.h"
#include "preproc.h"
#include "primary.h"
#include "stmt.h"
#include "sym.h"
#include "while.h"

/*
 *	statement parser
 *
 *	called whenever syntax requires a statement.  this routine
 *	performs that statement and returns a number telling which one
 *
 *	'func' is true if we require a "function_statement", which
 *	must be compound, and must contain "statement_list" (even if
 *	"declaration_list" is omitted)
 */

int statement (int func)
{
	if ((ch () == 0) & feof (input))
		return (0);
	lastst = 0;
	if (func)
	{
		if (match ("{")) {
			compound (YES);
			return (lastst);
		} else
			error ("function requires compound statement");
	}
	if (match ("{"))
		compound (NO);
	else
		stst ();
	return (lastst);
}

/*
 *	declaration
 */
int stdecl (void)
{
	if (amatch("register", 8))
		doldcls(DEFAUTO);
	else if (amatch("auto", 4))
		doldcls(DEFAUTO);
	else if (amatch("static", 6))
		doldcls(LSTATIC);
	else if (doldcls(AUTO)) ;
	else
		return (NO);
	return (YES);
}

int doldcls(int stclass)
{
	blanks();
	if (amatch("char", 4))
		declloc(CCHAR, stclass);
	else if (amatch("int", 3))
		declloc(CINT, stclass);
	else if (stclass == LSTATIC || stclass == DEFAUTO)
		declloc(CINT, stclass);
	else
		return(0);
	ns();
	return(1);
}


/*
 *	non-declaration statement
 */
void stst (void )
{
	if (amatch ("if", 2)) {
		doif ();
		lastst = STIF;
	} else if (amatch ("while", 5)) {
		dowhile ();
		lastst = STWHILE;
	} else if (amatch ("switch", 6)) {
		doswitch ();
		lastst = STSWITCH;
	} else if (amatch ("do", 2)) {
		dodo ();
		ns ();
		lastst = STDO;
	} else if (amatch ("for", 3)) {
		dofor ();
		lastst = STFOR;
	} else if (amatch ("return", 6)) {
		doreturn ();
		ns ();
		lastst = STRETURN;
	} else if (amatch ("break", 5)) {
		dobreak ();
		ns ();
		lastst = STBREAK;
	} else if (amatch ("continue", 8)) {
		docont ();
		ns ();
		lastst = STCONT;
	} else if (match (";"))
		;
	else if (amatch ("case", 4)) {
		docase ();
		lastst = statement (NO);
	} else if (amatch ("default", 7)) {
		dodefault ();
		lastst = statement (NO);
	} else if (match ("#asm")) {
		doasm ();
		lastst = STASM;
	} else if (match ("{"))
		compound (NO);
	else {
		expression (YES);
/*		if (match (":")) {
			dolabel ();
			lastst = statement (NO);
		} else {
*/			ns ();
			lastst = STEXP;
/*		}
*/	}
}

/*
 *	compound statement
 *
 *	allow any number of statements to fall between "{" and "}"
 *
 *	'func' is true if we are in a "function_statement", which
 *	must contain "statement_list"
 */
void compound (int func)
{
	int	decls;

	decls = YES;
	ncmp++;
	while (!match ("}")) {
		if (feof (input))
			return;
		if (decls) {
			if (!stdecl ()) {
				decls = NO;
				gtext ();
			}
		} else
			stst ();
	}
	ncmp--;
}

/*
 *	"if" statement
 */
void doif (void )
{
	int	fstkp, flab1, flab2;
	char	*flev;

	flev = locptr;
	fstkp = stkp;
	flab1 = getlabel ();
	test (flab1, FALSE);
	statement (NO);
	stkp = modstk (fstkp);
	locptr = flev;
	if (!amatch ("else", 4)) {
		gnlabel (flab1);
		return;
	}
	jump (flab2 = getlabel ());
	gnlabel (flab1);
	statement (NO);
	stkp = modstk (fstkp);
	locptr = flev;
	gnlabel (flab2);
}

/*
 *	"while" statement
 */
void dowhile (void )
{
	loop_expr_data_t ws;

	ws.symbol = locptr;
	ws.sp = stkp;
	ws.type = WSWHILE;
	ws.casep_test = getlabel ();
	ws.exit = getlabel ();
	addwhile (&ws);
	gnlabel (ws.casep_test);
	test (ws.exit, FALSE);
	statement (NO);
	jump (ws.casep_test);
	gnlabel (ws.exit);
	locptr = ws.symbol;
	stkp = modstk (ws.sp);
	delwhile ();
}

/*
 *	"do" statement
 */
void dodo (void )
{
	loop_expr_data_t ws;

	ws.symbol = locptr;
	ws.sp = stkp;
	ws.type = WSDO;
	ws.body_tab = getlabel ();
	ws.casep_test = getlabel ();
	ws.exit = getlabel ();

	addwhile (&ws);
	gnlabel (ws.body_tab);
	statement (NO);
	if (!match ("while")) {
		error ("missing while");
		return;
	}
	gnlabel (ws.casep_test);
	test (ws.body_tab, TRUE);
	gnlabel (ws.exit);
	locptr = ws.symbol;
	stkp = modstk (ws.sp);
	delwhile ();
}

/*
 *	"for" statement
 */
void dofor (void )
{
	loop_expr_data_t *pws;
	loop_expr_data_t ws;

	ws.symbol = locptr;
	ws.sp = stkp;
	ws.type = WSFOR;
	ws.casep_test = getlabel ();
	ws.incr_def = getlabel ();
	ws.body_tab = getlabel ();
	ws.exit = getlabel ();

	addwhile (&ws);
	pws = readwhile ();
	needbrack ("(");
	if (!match (";")) {
		expression (YES);
		ns ();
	}
	gnlabel (pws->casep_test);
	if (!match (";")) {
		expression (YES);
		testjump (pws->body_tab, TRUE);
		jump (pws->exit);
		ns ();
	} else
		pws->casep_test = pws->body_tab;
	gnlabel (pws->incr_def);
	if (!match (")")) {
		expression (YES);
		needbrack (")");
		jump (pws->casep_test);
	} else
		pws->incr_def = pws->casep_test;
	gnlabel (pws->body_tab);
	statement (NO);
	jump (pws->incr_def);
	gnlabel (pws->exit);
	locptr = pws->symbol;
	stkp = modstk (pws->sp);
	delwhile ();
}

/*
 *	"switch" statement
 */
void doswitch (void )
{
	loop_expr_data_t	ws;
	loop_expr_data_t	*ptr;

	ws.symbol = locptr;
	ws.sp = stkp;
	ws.type = WSSWITCH;
	ws.casep_test = swstp;
	ws.body_tab = getlabel ();
	ws.incr_def = ws.exit = getlabel ();
	addwhile (&ws);
	immed (T_LABEL, (DATA){.integer=ws.body_tab});
	gpush ();
	needbrack ("(");
	expression (YES);
	needbrack (")");
	stkp = stkp + INTSIZE;  /* '?case' will adjust the stack */
	gjcase ();
	statement (NO);
	ptr = readswitch ();
	jump (ptr->exit);
	dumpsw (ptr);
	gnlabel (ptr->exit);
	locptr = ptr->symbol;
	stkp = modstk (ptr->sp);
	swstp = ptr->casep_test;
	delwhile ();
}

/*
 *	"case" label
 */
void docase (void )
{
	int	val;

	val = 0;
	if (readswitch ()) {
		if (!number (&val))
			if (!pstr (&val))
				error ("bad case label");
		addcase (val);
		if (!match (":"))
			error ("missing colon");
	} else
		error ("no active switch");
}

/*
 *	"default" label
 */
void dodefault (void )
{
	int	lab;
	loop_expr_data_t * ptr;

	ptr = readswitch ();
	if (ptr) {
		ptr->incr_def = lab = getlabel ();
		gnlabel (lab);
		if (!match (":"))
			error ("missing colon");
	} else
		error ("no active switch");
}

/*
 *	"return" statement
 */
void doreturn (void )
{
	if (endst () == 0)
		expression (YES);
	jump(fexitlab);
}

/*
 *	"break" statement
 */
void dobreak (void )
{
	loop_expr_data_t	*ptr;

	if ((ptr = readwhile ()) == 0)
		return;
	modstk (ptr->sp);
	jump (ptr->exit);
}

/*
 *	"continue" statement
 */
void docont (void )
{
	loop_expr_data_t	*ptr;

	if ((ptr = findwhile ()) == 0)
		return;
	modstk (ptr->sp);
	if (ptr->type == WSFOR)
		jump (ptr->incr_def);
        	else
		jump (ptr->casep_test);
}

/*
 *	dump switch table
 */
void dumpsw (loop_expr_data_t *ws)
/*int	ws[];*/
{
	int	i,j;

//	gdata ();
	gnlabel (ws->body_tab);
	if (ws->casep_test != swstp) {
		j = ws->casep_test;
		while (j < swstp) {
			defword ();
			i = 4;
			while (i--) {
				outdec (swstcase[j]);
				outbyte (',');
				outlabel (swstlab[j++]);
				if ((i == 0) | (j >= swstp)) {
					nl ();
					break;
				}
				outbyte (',');
			}
		}
	}
	defword ();
	outlabel (ws->incr_def);
	outstr (",0");
	nl ();
//	gtext ();
}

void test (int label, int ft)
/* int	label,
	ft; */
{
	needbrack ("(");
	expression (YES);
	needbrack (")");
	testjump (label, ft);
}

