; Thx to cgfm 2
;  Use a look up table for frequency
;      9x12 high and low
;  Reorganize instrument data
;      length     : inst0 inst1 ... instN
;      repeat idx : inst0 inst1 ... instN
;      data inst0 : byte0 byte1 ....
;      data inst1 : byte0 byte1 ....
;      data inst2 : byte0 byte1 ....
;      data inst3 : byte0 byte1 ....
;
;      so that updateSequence will look like:
;      lda     [length]
;
;------------------------------------------------------------------------
; Macros
;------------------------------------------------------------------------
;
; Map and update sequence pointer
; \1 sequence name
; \2 jump point to sequence end
updateSequencePtr   .macro
    ldy    <instrument+\1,X
    lda    <instrument+\1+6,X
    cmp    sequenceLength,Y              ; Check if we reached sequence end
    bne    .updateSequencePtr_\@
        lda   sequenceRepeatIndex,Y    ; Reset sequence index (loop)
        cmp   #128                     ; There's no loop if the reset index is
        bne   .updateSequencePtr_\@    ; equal to 128
        jmp   \2
.updateSequencePtr_\@:
    tay
    inc    A
    sta    <instrument+\1+6,X
                    .endm

;
; Set frequency according to tone and octave
; Input:
;	\1 : word containing tone and octave (zp)
;	\2 : word for frequency (zp)
setFrequency .macro
    ; Compute frequency
    ldy    <\1

    ; We don't have to map freqTables as it's in the prg bank
    lda    freqTableLow,y
    sta    <\2
    lda    freqTableHigh,y

    ; So we need to left shift it.
    ldy    <\1+1
    beq    .freq_nothing\@
.freq_loop_\@:
    lsr    A
    ror    <\2
    dey
    bne    .freq_loop_\@
.freq_nothing\@:
    sta    <\2+1
           .endm

;
; Compute frequency table index from octave and tone
; Input:
;       \1 : octave
;       \2 : tone
computeIndexFromNote .macro
    lda #\2
    sta <note+tone
    lda #\1
    sta <note+octave
    asl A                  ; octave * 12 = ((octave * 2 + octave) * 4
    clc
    adc #\1
    asl A
    asl A
    adc #\2                ; + tone
    sta <note+index
                     .endm
;------------------------------------------------------------------------
; Internal variables (declared in zp)
;------------------------------------------------------------------------
         .zp
freq_ptr        ds    2                                      ; pointer to frequency table
instrument_ptr  ds    2                                      ; pointer to instrument data
channel         ds    1                                      ; current channel in use
update          ds    1                                      ; update flag (frequency, volume,..)
note            ds    NOTE_SIZE                              ; current note (tone, octave)
delim           ds    2                                      ;<<<< DEBUG PURPOSE
instrument                                                   ; instruments (one per channel)
instrument_flag        ds 6                                  ; > flag
                                                             ;   bit 0 : arpeggio
                                                             ;       1 : left volume
                                                             ;       2 : right volume
                                                             ;       3 : pitch

left_volume_ptr_lo     ds 6                                  ; > left volume sequence pointer
left_volume_ptr_hi     ds 6
left_volume_sequence   ds 6                                  ; > left volume sequence index
left_volume_offset     ds 6                                  ; > left volume sequence offset
right_volume_ptr_lo    ds 6                                  ; > right volume sequence pointer
right_volume_ptr_hi    ds 6
right_volume_sequence  ds 6                                  ; > right volume sequence index
right_volume_offset    ds 6                                  ; > right volume sequence offset
arpeggio_ptr_lo        ds 6                                  ; > arpeggio sequence pointer
arpeggio_ptr_hi        ds 6
arpeggio_sequence      ds 6                                  ; > arpeggio sequence index
arpeggio_offset        ds 6                                  ; > arpeggio sequence offset
pitch_ptr_lo           ds 6                                  ; > pitch sequence pointer
pitch_ptr_hi           ds 6
pitch_sequence         ds 6                                  ; > pitch sequence index
pitch_offset           ds 6                                  ; > pitch sequence offset

         .code
;------------------------------------------------------------------------
; Frequency table
;------------------------------------------------------------------------
; Give an octave and a note, we can compute the associated frequency using the formula:
;     f = ref * 2^((octave-4) + ((tone-10)/12))
; On pcengine, the frequency in stored on 12 bit:
;     d = 7160000 / (2 * 32 * f)
; So after some computations we have:
;     d = (7160000 / (ref * 2^((14+tone)/12))) / 2^octave
; We can compute a table containing all twelves notes for the first octave
; and then shift it to the right times the octave.
; The lowest note playable on pcengine is C-1 ( :) ) (remember the frequency is encoded on 12bits).
; So the highest possible note will be A9.
; For simplicity reason, we splited the frequency table into 2 tables. On for the hight bytes and
; the other for the low bytes
  .include "frequency.inc"

;------------------------------------------------------------------------
; Music routine
;       X : channel
;       Y : instrument
;------------------------------------------------------------------------

msx_update:

    ; #### TODO : DO THIS WHEN WE PLAY A NEW NOTE [BEGIN]
    ldx    #0                ; #### test
    ldy    #2                ; #### test
    bne    sequenceUpdate
    lda    <instrument_flag,X
    sta    <_al
    jmp    sequenceUpdateEnd
sequenceUpdate:
    ; Map instrument data
    lda    #bank(instruments)
    tam    #SOUND_MPR

    ; Get instrument pointer
    lda   instrumentPointerLow,Y
    sta    <__ptr
    lda   instrumentPointerHigh,Y
    sta    <__ptr+1

    ; flag
    lda   [__ptr]
    sta   <instrument_flag,X
    sta   <_al

    ; Set sequence pointers
setVolumeLeft
    bbr0   <_al, setVolumeRight
    stz    <left_volume_offset,X                  ; reset sequence offset
    ldy    #1
    lda    [__ptr],Y                              ; get sequence index
    sta    <left_volume_sequence,X
    tay
    lda    sequenceDataPointerLow,Y               ; get sequence pointer
    sta    <left_volume_ptr_lo,X
    lda    sequenceDataPointerHigh,Y
    sta    <left_volume_ptr_hi,X

setVolumeRight:
    bbr1   <_al, setArpeggio
    stz    <right_volume_offset,X                 ; reset sequence offset
    ldy    #2
    lda    [__ptr],Y                              ; get sequence index
    sta    <right_volume_sequence,X
    tay
    lda    sequenceDataPointerLow,Y               ; get sequence pointer
    sta    <right_volume_ptr_lo,X
    lda    sequenceDataPointerHigh,Y
    sta    <right_volume_ptr_hi,X

setArpeggio:
    bbr2   <_al, setPitch
    stz    <arpeggio_offset,X                    ; reset sequence offset
    ldy    #3
    lda    [__ptr],Y                              ; get sequence index
    sta    <arpeggio_sequence,X
    tay
    lda    sequenceDataPointerLow,Y               ; get sequence pointer
    sta    <arpeggio_ptr_lo,X
    lda    sequenceDataPointerHigh,Y
    sta    <arpeggio_ptr_hi,X

setPitch:
    bbr3   <_al, sequenceUpdateEnd
    stz    <pitch_offset,X                        ; reset sequence offset
    ldy    #4
    lda    [__ptr],Y                              ; get sequence index
    sta    <pitch_sequence,X
    tay
    lda    sequenceDataPointerLow,Y               ; get sequence pointer
    sta    <pitch_ptr_lo,X
    lda    sequenceDataPointerHigh,Y
    sta    <pitch_ptr_hi,X

sequenceUpdateEnd:
    ; #### TODO : DO THIS WHEN WE PLAY A NEW NOTE [END]

    ; Flag telling what to update
    ; bit #0 : Volume
    ;      1 : Frequency
    stz    <update

    ;------------------------------------
    ; Volume [begin]
    ;------------------------------------
    lda    <_al
    and    #$06
    beq    volume_end                  ; Are you going to update volume?
    smb0   <update                     ; update volume
volume_begin:
volume_left:
    bbr0   <_al, volume_right
    updateSequencePtr left_volume_sequence_idx, volume_right
    lda    <left_volume_ptr_lo,X
    sta    <__ptr
    lda    <left_volume_ptr_hi,X
    sta    <__ptr+1
    lda    [__ptr],Y
    asl    A
    asl    A
    asl    A
    asl    A
    sta    <__tmp
    lda    <note+volume
    and    #$0F
    ora    <__tmp
    sta    <note+volume                ; volume = (volume & 0x0f) | (sequence[idx] << 4)

volume_right:
    bbr1   <_al, volume_end
    updateSequencePtr right_volume_sequence_idx, volume_end
    lda    <right_volume_ptr_lo,X
    sta    <__ptr
    lda    <right_volume_ptr_hi,X
    sta    <__ptr+1
    lda    <note+volume
    and    #$F0                        ; volume = (volume & 0xf0) | sequence[idx]
    ora    [__ptr],Y
    sta    <note+volume
    ;------------------------------------
    ; volume [end]
    ;------------------------------------
volume_end:

    ;------------------------------------
    ; Arpeggio [begin]
    ;------------------------------------
    bbr2    <_al, arpeggio_end
    updateSequencePtr arpeggio_sequence_idx, arpeggio_end
    lda    <arpeggio_ptr_lo,X
    sta    <__ptr
    lda    <arpeggio_ptr_hi,X
    sta    <__ptr+1
    smb1   <update                      ; update frequency
arpeggio_begin:
    clc
    lda    <note+index
    adc    [__ptr],Y                    ; tone += sequence[idx]
    tay
    bmi    .arpeggio_negative           ; tone < 0 ?
    cmp    #108
    bmi    .arpeggio_update             ; tone < 9*12 ?
    ldy    #107
    bra    .arpeggio_update
.arpeggio_negative:
    cly
.arpeggio_update:
    lda    freqTableLo,Y
    sta    <note+frequency_l
    lda    freqTableHi,Y
    sta    <note+frequency_h

    ;------------------------------------
    ; Arpeggio [end]
    ;------------------------------------
arpeggio_end:

    ;------------------------------------
    ; Pitch [begin]
    ;------------------------------------
    bbr3    <_al, pitch_end
    updateSequencePtr pitch_sequence_idx, pitch_end
    lda    <pitch_ptr_lo,X
    sta    <__ptr
    lda    <pitch_ptr_hi,X
    sta    <__ptr+1
    smb1   <update                    ; update frequency
pitch_begin:
    clc
    lda    [__ptr],Y
    bmi    pitch_neg
    ; positive case
        adc    <note+frequency_l
        sta    <note+frequency_l
        cla
        adc    <note+frequency_h
        sta    <note+frequency_h         ; frequency += sequence[idx]
        jmp    pitch_end
    ; negative case
pitch_neg:
        clc
        adc    <note+frequency_l
        sta    <note+frequency_l
        lda    <note+frequency_h
        sbc    #0
        sta    <note+frequency_h
    ;------------------------------------
    ; Pitch [end]
    ;------------------------------------
pitch_end:

    ; set volume
    bbr0   <update,.mute             ; update volume if bit #0 is set
    lda    <note+volume
    sta    psg_pan
.mute:

    ; play note
    bbr1   <update,.no_note          ; update frequency if bit #1 is set
    lda    <note+frequency_l
    sta    psg_freqlo
    lda    <note+frequency_h
    sta    psg_freqhi
.no_note:

    rts
