       	.include "equ.inc"
	.include "macro.inc"
	.include "vdc.inc"

	.zp

__tmp          ds    2
__ptr          ds    2

irq1_jmp       ds    2
hsync_hook     ds    2
vsync_hook     ds    2

	.code

;------------------------------------------------------------------------
; Interrupt vectors
;------------------------------------------------------------------------
	.code
	.bank 0
        .org $E000

        .include "gfx.asm"

vdcInitTable:
;       reg  low  hi
    .db $07, $00, $00 ; background x-scroll register
    .db $08, $00, $00 ; background y-scroll register
    .db $09, $10, $00 ; background map size
    .db $0A, $02, $02 ; horizontal period register
    .db $0B, $1F, $04 ; horizontal display register
    .db $0C, $02, $17 ; vertical sync register
    .db $0D, $DF, $00 ; vertical display register
    .db $0E, $0C, $00 ; vertical display position end register

;------------------------------------------------------------------------
; Reset routine
;------------------------------------------------------------------------
irq_reset:

    sei                                ; disable interrupts
    csh                                ; select the 7.16 MHz clock
    cld                                ; clear the decimal flag
    ldx    #$FF                        ; initialize the stack pointer
    txs
    lda    #$FF                        ; map the I/O bank in the first page
    tam    #0
    lda    #$F8                        ; and the RAM bank in the second page
    tam    #1
    stz    $2000                       ; clear all the RAM
    tii    $2000,$2001,$1FFF

    lda    #%11111101
    sta    irq_disable
    stz    irq_status

    ; Set tia ram function
    lda   #$E3 ; tia
    sta   _ram_hdwr_tia
    lda   #$60 ; rts
    sta   _ram_hdwr_tia_rts

    lda    #1
    sta    timer_ctrl                   ; disable timer

    st0    #$05                         ; set vdc control register
    st1    #$00                         ; disable vdc interupts
    st2    #$00                         ; sprite and bg are disabled

    ; initialize vdc
    lda    #low(vdcInitTable)
    sta    <__tmp
    lda    #high(vdcInitTable)
    sta    <__tmp+1

    cly
.l0:
        lda    [__tmp],Y
        sta    videoport
        iny
        lda    [__tmp],Y
        sta    video_data_l
        iny
        lda    [__tmp],Y
        sta    video_data_h
        iny
        cpy    #24
        bne    .l0

    ; disable sound
    lda    #$06                         ; pcengine has 6 channels
.l1
        dec    A                        ; chose channel
        sta    psg_ch
        stz    psg_ctrl                 ; disable channel
        bne    .l1

    rti

;------------------------------------------------------------------------
; Timer interrupt
;------------------------------------------------------------------------
irq_timer:

    stz    irq_status                  ; acknowledge interrupt

    rti

;------------------------------------------------------------------------
; NMI interrupt handler
;------------------------------------------------------------------------
irq_nmi:
    rti

;------------------------------------------------------------------------
; IRQ 1 interrupt handler
;------------------------------------------------------------------------
irq_1:
    bbs1    <irq_m,user_irq1              ; jump to the user irq1 vector if bit set
    pha                                   ; save registers
    phx
    phy

    lda   video_reg		          ; get VDC status register
    sta  <vdc_sr

    ; vsync and hsync are mixed
    ; the only way to separate them is to check the value
    ; of the vdc status register
.vsync:                                   ; vsync interrupt
    ; if bit 5 is set we are on vsync
    bbr5   <vdc_sr,.hsync

    st0    #5                             ; update display control (bg/sp)
    lda    <vdc_crl                       ; vdc control register
    sta    video_data_l

    bbs5 <irq_m, .l3                      ; do we want standard vsync handling ?
	jsr    _vsync_handler             ; standard vsync handler
.l3
    bbr4   <irq_m, .hsync
        jsr    user_vsync                 ; let's run user's vsync hook

                                          ; no standard hsync handling here :D
.hsync:                                   ; hsync interrupt
    bbr2    <vdc_sr, .exit                ; check if hsync bit is set in vdc status register
    bbr6    <irq_m, .exit
        jsr    user_hsync                 ; let's run user's hsync hook
.exit:

    ; restore registers
    lda    <vdc_reg
    sta    video_reg
    ply
    plx
    pla

    rti

user_irq1:
    jmp   [irq1_jmp]
user_hsync:
    jmp   [hsync_hook]
user_vsync:
    jmp   [vsync_hook]

;--------------------------------
; Standard vsync handler
;--------------------------------
_vsync_handler:
    rts

;------------------------------------------------------------------------
; IRQ 2 interrupt handler
;------------------------------------------------------------------------
irq_2:
    rti

;------------------------------------------------------------------------
; Vector table
;------------------------------------------------------------------------
        .org $FFF6

        .dw irq_2                    ; irq 2
        .dw irq_1                    ; irq 1
        .dw irq_timer                ; timer
        .dw irq_nmi                  ; nmi
        .dw irq_reset                ; reset

